# http2-wrapper
> Use HTTP2 the same way like HTTP1

[![Build Status](https://travis-ci.org/szmarczak/http2-wrapper.svg?branch=master)](https://travis-ci.org/szmarczak/http2-wrapper)
[![Coverage Status](https://coveralls.io/repos/github/szmarczak/http2-wrapper/badge.svg?branch=master)](https://coveralls.io/github/szmarczak/http2-wrapper?branch=master)
[![npm](https://img.shields.io/npm/dm/http2-wrapper.svg)](https://www.npmjs.com/package/http2-wrapper)
[![install size](https://packagephobia.now.sh/badge?p=http2-wrapper)](https://packagephobia.now.sh/result?p=http2-wrapper)

This package was created for the purpose of supporting HTTP2 without rewriting your code.<br>
I recommend adapting to the `http2` module if possible, because it's much simpler to use and has many cool features.

**Tip**: `http2-wrapper` is very useful when you rely on other modules that use the HTTP1 API and you want to support HTTP2.

## Usage
```js
'use strict';
const http2 = require('http2-wrapper');

const options = {
	hostname: 'nghttp2.org',
	protocol: 'https:',
	path: '/httpbin/post',
	method: 'POST',
	headers: {
		'content-length': 6
	}
};

const req = http2.request(options, res => {
	console.log('statusCode:', res.statusCode);
	console.log('headers:', res.headers);

	const body = [];
	res.on('data', chunk => {
		body.push(chunk);
	});
	res.on('end', () => {
		console.log('body:', Buffer.concat(body).toString());
	});
});

req.on('error', e => console.error(e));

req.write('123');
req.end('456');

// statusCode: 200
// headers: { ':status': 200,
//   date: 'Sat, 11 Aug 2018 09:37:41 GMT',
//   'content-type': 'application/json',
//   'content-length': '264',
//   'access-control-allow-origin': '*',
//   'access-control-allow-credentials': 'true',
//   'x-backend-header-rtt': '0.002997',
//   'strict-transport-security': 'max-age=31536000',
//   server: 'nghttpx',
//   via: '1.1 nghttpx',
//   'x-frame-options': 'SAMEORIGIN',
//   'x-xss-protection': '1; mode=block',
//   'x-content-type-options': 'nosniff' }
// body: {
//   "args": {},
//   "data": "123456",
//   "files": {},
//   "form": {},
//   "headers": {
//     "Content-Length": "6",
//     "Host": "nghttp2.org:443",
//     "Via": "2 nghttpx"
//   },
//   "json": 123456,
//   "origin": "xxx.xxx.xxx.xxx",
//   "url": "https://nghttp2.org:443/httpbin/post"
// }
```

## Differences between the HTTP1 API

This wrapper may work a bit different than the HTTP1 API:

 - the `upgrade` event is not supported (yet),
 - the `HTTP2ClientRequest` class has no `setSocketKeepAlive` function,
 - `close` events inform that the underlying request is no longer usable (instead of: the socket has been closed),
 - there's a `session` option (accepts [`ClientHttp2Session`](https://nodejs.org/api/http2.html#http2_http2_connect_authority_options_listener)) instead of the `agent` option.
 
### HTTP2IncomingMessage

 * [x] Event: 'aborted'
 * [x] Event: 'close'
 * [x] message.destroy([error])
 * [x] message.headers
 * [x] message.httpVersion
 * [x] message.rawHeaders
 * [x] message.rawTrailers
 * [x] message.setTimeout(msecs, callback)
 * [x] message.socket
 * [x] message.statusCode
 * [x] message.statusMessage
 * [x] message.trailers

### HTTP2ClientRequest
 * [x] Event: 'abort'
 * [x] Event: 'connect'
 * [x] Event: 'continue'
 * [x] Event: 'information'
 * [x] Event: 'response'
 * [x] Event: 'socket'
 * [x] Event: 'timeout'
 * [ ] Event: 'upgrade'
 * [x] request.abort()
 * [x] request.aborted
 * [x] request.connection
 * [x] request.end([data[, encoding]][, callback])
 * [x] request.flushHeaders()
 * [x] request.getHeader(name)
 * [x] request.removeHeader(name)
 * [x] request.setHeader(name, value)
 * [x] request.setNoDelay([noDelay])
 * [ ] ~~request.setSocketKeepAlive([enable][, initialDelay])~~
 * [x] request.setTimeout(timeout[, callback])
 * [x] request.socket
 * [x] request.write(chunk[, encoding][, callback])

## License

MIT
