'use strict';
const http2 = require('http2');
const {Writable} = require('stream');
const deferToConnect = require('defer-to-connect');
const HTTP2IncomingMessage = require('./HTTP2IncomingMessage');
const proxySocket = require('./proxySocket');
const {
	ERR_HTTP2_HEADERS_SENT,
	ERR_INVALID_HTTP_TOKEN,
	ERR_HTTP2_INVALID_HEADER_VALUE,
	SOCKET_HANG_UP
} = require('./errors');

const {
	HTTP2_HEADER_STATUS,
	HTTP2_HEADER_METHOD,
	HTTP2_HEADER_PATH,
	NGHTTP2_CANCEL
} = http2.constants;

const kBody = Symbol('body');
const kHeaders = Symbol('headers');
const kCustomSession = Symbol('customSession');

class HTTP2ClientRequest extends Writable {
	constructor(options, cb) {
		super();

		if (typeof options === 'string') {
			options = new URL(options);
		} else {
			options = {...options};

			// Accepts custom authority
			if (options.authority) {
				const parsed = new URL(options.authority);
				options.port = parsed.port;
				options.hostname = parsed.hostname;
				options.host = parsed.hostname;
			} else {
				// We should normalize options.port on our own,
				// because the TLS API and HTTP2 API normalize it in a different way.
				options.port = options.port ? options.port : (options.protocol === 'http:' ? 80 : 443);
			}
		}

		const self = this;

		this[kBody] = [];
		this[kHeaders] = {};

		if (options.session) {
			// Using a custom HTTP2 session, so we don't need to care about closing it later.
			this[kCustomSession] = true;
			this.session = options.session;
		} else {
			this[kCustomSession] = false;

			this.session = http2.connect(options.authority || options, options);
		}

		// Workaround for #2 - reproxy the original socket,
		// because the session will delete the Symbol(kSocket) property after
		// it has been destroyed. We need to save it for later.
		this.socket = proxySocket(this.session, this.session.socket.ref());
		this.connection = this.socket;

		// Forwards errors to this instance.
		this.session.on('error', err => self.emit('error', err));

		this.socket.on('close', hadError => {
			if (!self.res && !self[kCustomSession] && !hadError && !self.socket._hadError) {
				// Request has been ended with no response,
				// so throw the `socket hang up` error.
				self.socket._hadError = true;
				self.emit('error', new SOCKET_HANG_UP());
			}
		});

		this.method = options.method || 'GET';
		this.path = options.path || '/';

		if (cb) {
			this.once('response', cb);
		}

		// This gets called on `request.end()`.
		this.once('finish', () => {
			if (this.session.destroyed) {
				// The session has been destroyed, it's no longer usable.
				return;
			}

			if (!this._request) {
				// Haven't made a real request yet, we need to make one.
				this.flushHeaders();
			}

			// Writes queued data.
			for (const element of this[kBody]) {
				this._request.write(element);
			}

			this._request.end();
		});

		this.res = null;
		this.aborted = false;

		if (options.headers) {
			// Applies provided headers
			for (const [header, value] of Object.entries(options.headers)) {
				this.setHeader(header, value);
			}
		}

		// Move the `socket` event to next tick, so the user can
		// attach the listener before the event has been emitted.
		process.nextTick(() => self.emit('socket', this.socket));
	}

	_write(chunk, enc, next) {
		if (this._request && this._request.writable) {
			this._request.write(chunk, enc);
		} else {
			this[kBody].push(Buffer.from(chunk, enc));
		}
		next();
	}

	abort() {
		if (!this.aborted) {
			process.nextTick(self => self.emit('abort'), this);
		}
		this.aborted = true;

		if (this.res) {
			this.res._dump();
		}

		if (this._request) {
			this._request.close(NGHTTP2_CANCEL);
		} else if (!this[kCustomSession]) {
			this.session.close();
		}
	}

	flushHeaders() {
		if (!this._request) {
			const self = this;

			// Makes the real request.
			const headers = {
				[HTTP2_HEADER_METHOD]: this.method,
				...this[kHeaders]
			};

			// Update the `method` property in case users changes it
			this.method = headers[HTTP2_HEADER_METHOD].toUpperCase();

			// Must not specify the `:path` and `:scheme` headers for `CONNECT` requests.
			if (this.method !== 'CONNECT') {
				headers[HTTP2_HEADER_PATH] = this.path;
			}

			this._request = this.session.request(headers);

			const proxyEvent = e => this._request.on(e, (...args) => self.emit(e, ...args));

			// Proxy these events, we don't need them to implement them.
			['timeout', 'continue'].forEach(proxyEvent);

			if (self.method !== 'CONNECT') {
				// Don't emit the same error twice,
				// because we listen for them on line 60.
				this._request.on('error', err => {
					if (!err.message.includes('caused by')) {
						self.emit('error', err);
					}
				});
			}

			// This event tells we are ready to listen for the data.
			this._request.on('response', (headers, flags, rawHeaders) => {
				self.res = new HTTP2IncomingMessage(self, {headers, flags, rawHeaders});

				if (self.method === 'CONNECT') {
					// The HTTP1 API says the socket is detached here,
					// but we can't do that so we pass the original HTTP2 request.
					if (self.emit('connect', self.res, self._request, Buffer.alloc(0))) {
						process.nextTick(() => self.res.push(null));
					} else {
						// No listeners attached, destroy the original request.
						self._request.destroy();
					}
				} else {
					// Forwards the `pause`, `resume`, `end` and `close` events.
					self.res.on('pause', () => self._request.pause());
					self.res.on('resume', () => self._request.resume());

					self._request.on('close', () => self.res.emit('close'));
					self._request.on('end', () => self.res.push(null));

					if (!self.emit('response', self.res)) {
						// No listeners attached, dump the response.
						self.res._dump();
					}
				}
			});

			/* istanbul ignore next */
			this._request.on('headers', headers => {
				if (headers[HTTP2_HEADER_STATUS] >= 102) {
					self.emit('information', {statusCode: headers[HTTP2_HEADER_STATUS]});
				}
			});

			this._request.on('trailers', (trailers, flags, rawTrailers) => {
				// Assigns trailers to the response object.
				Object.assign(self.res.trailers, trailers);
				Object.assign(self.res.rawTrailers, rawTrailers);
			});

			if (!this[kCustomSession]) {
				// We own this session, so we need to close it (usually).
				this._request.once('close', () => self.session.close());
			}

			this.emit('flushedHeaders');
		}
	}

	getHeader(name) {
		return this[kHeaders][name];
	}

	get headersSent() {
		if (this._request) {
			return Boolean(this._request.sentHeaders);
		}

		return false;
	}

	removeHeader(name) {
		if (this.headersSent) {
			throw new ERR_HTTP2_HEADERS_SENT('remove');
		}
		if (name === '' || typeof name !== 'string') {
			throw new ERR_INVALID_HTTP_TOKEN('Header name', name);
		}
		name = name.trim().toLowerCase();

		delete this[kHeaders][name];
	}

	setHeader(name, value) {
		if (this.headersSent) {
			throw new ERR_HTTP2_HEADERS_SENT('set');
		}
		if (name === '' || typeof name !== 'string') {
			throw new ERR_INVALID_HTTP_TOKEN('Header name', name);
		}
		name = name.trim().toLowerCase();
		if (value === undefined || value === null) {
			throw new ERR_HTTP2_INVALID_HEADER_VALUE(value, name);
		}

		this[kHeaders][name] = value;
	}

	setNoDelay(noDelay) {
		deferToConnect(this.socket, 'setNoDelay', noDelay);
	}

	setTimeout(ms, cb) {
		if (this._request) {
			this._request.setTimeout(ms, cb);
		} else {
			const self = this;

			// The request hasn't been made yet, need to listen for it.
			this.once('flushedHeaders', () => {
				self._request.setTimeout(ms, cb);
			});
		}

		return this;
	}
}

module.exports = HTTP2ClientRequest;
